\ TSEPOS 193 1 1 \
\ TSECFGEXPR    '^[ \t]*{\(\ \**}|{code}|{:}'
\ TSECFGOPTS    'ix'

\ T             title
\ P             begin paragraph (ends after any non * line)
\ H             a general heading
\ S             Section sub-title
\ D             Definition, next word is title, all following text
\               and * lines are descriptions
\ N             Forced newline
\ C             Description comment for Forth word
\ E             An Edit Comment line

\ =======================================================================
( **                                                                    )
( *T i82527 CAN Controller Device Driver                                )
( **                                                                    )
\ =======================================================================


\ =======================================================================
( **                                                                    )
( *S Defines and Constants                                              )
( **                                                                    )
( *H Register Name Contants                                             )
( *D CAN_CONTROL                Control Register                        )
( *D CAN_STATUS                 Status Register                         )
( *D CAN_CPUINT                 CPU Interface Reg                       )
( *D CAN_BUSCONF                Bus Config Reg                          )
( *D CAN_BTR0                   Bit Timing Register #0                  )
( *D CAN_BTR1                   Bit Timing Register #1                  )
( *D CAN_GMS                    Global Mask - Standard                  )
( *D CAN_GME                    Global Mask - Extended                  )
( *D CAN_INTSRC                 Interrupt Source Specifier              )
( *D CAN_P2CONF                 Configure IO Port #2                    )
( *D CAN_P2IN                   Port #2 input reg                       )
( *D CAN_P2OUT                  Port #2 output reg                      )
( *D MES_CTR0                   Message Control Reg #0                  )
( *D MES_CTR1                   Message Control Reg #1                  )
( *D MES_ARBI0                  Message Arb #0                          )
( *D MES_ARBI1                  Message Arb #1                          )
( *D MES_ARBI2                  Message Arb #2                          )
( *D MES_ARBI3                  Message Arb #3                          )
( *D MES_CONF                   Message Configuration                   )
( *D MES_DATA                   start of data packet                    )
( **                                                                    )
\ =======================================================================

$000            constant        CAN_CONTROL
$001            constant        CAN_STATUS
$002            constant        CAN_CPUINT
$02F            constant        CAN_BUSCONF
$03F            constant        CAN_BTR0
$04F            constant        CAN_BTR1
$006            constant        CAN_GMS
$008            constant        CAN_GME
$05F            constant        CAN_INTSRC
$0AF            constant        CAN_P2CONF
$0CF            constant        CAN_P2IN
$0EF            constant        CAN_P2OUT
$000            constant        MES_CTR0
$001            constant        MES_CTR1
$002            constant        MES_ARBI0
$003            constant        MES_ARBI1
$004            constant        MES_ARBI2
$005            constant        MES_ARBI3
$006            constant        MES_CONF
$007            constant        MES_DATA

\ =======================================================================
( **                                                                    )
( *S Hardware Interface Layer                                           )
( **                                                                    )
( *P The hardware access layer consists of a set of DEFERed words which )
( ** are patched in a HAL at a layer date                               )
( **                                                                    )
\ =======================================================================

variable        CAN_BASE                \ The hardware base address

defer           CAN-INIT
defer           CAN!
defer           CAN@
defer           CAN-FLUSH

\ =======================================================================
( **                                                                    )
( *S Diagnostic Words                                                   )
( **                                                                    )
\ =======================================================================

: RegDump               \ --
( ** DUMP i82527 registers                                              )
  base @ >r hex
  256 0
  do

    cr i s>d <# # # # # ascii . hold # # # # #> type
    2 spaces

    8 0 do
      j i + CAN@ s>d <# # # #> type
      space
    loop
    space
    16 8 do
      j i + CAN@ s>d <# # # #> type
      space
    loop

  $10
  +loop
  r> base !
;

\ =======================================================================
( **                                                                    )
( *S Utility Words                                                      )
( **                                                                    )
\ =======================================================================

: ConfigureMessage      { *data index | reg# -- }
( ** Set up one of the CAN message blocks from a memory data table      )
  index $10 * -> reg#
  $0F 0 do
    *data i + c@
    reg# i + CAN!
  loop
  can-flush
;

: SetMessageData        { mess# data len | -- }
( ** Set up the data portion of a CAN message block from memory         )

  $FB mess# $10 *
  MES_CTR1 + CAN!                         \ tell CAN we are updating message #

  data
  mess# $10 * MES_DATA +
  len 0 ?do                             \ src dst
    2dup swap c@
    swap CAN!
    1+ swap 1+ swap
  loop
  2drop

  $F6 mess# $10 *
  MES_CTR1 + CAN!                         \ tell CAN we have updated message #
  can-flush
;

: GetMessageData        { mess# buff len | -- }
( ** Read a CAN message block's data portion to a memory buffer.        )
  mess# $10 * MES_DATA +
  buff
  len 0 ?do
    2dup swap CAN@
    swap c!
    1+ swap 1+ swap
  loop
  2drop

  $FD mess# $10 * MES_CTR1 + CAN!       \ clear new data
  can-flush
;

: CanInt@               \ -- value
( ** Retrieve the i82527 Interrupt Source register.                     )
  CAN_INTSRC CAN@
;

create  defaultcan      $55 c, $55 c, $00 c, $00 c, $00 c, $00 c,
                        $00 c,
                        $00 c, $00 c, $00 c, $00 c,
                        $00 c, $00 c, $00 c, $00 c,

: PowerOnConfigure      \ --
( ** Perform MPE PowerUp configuration of i82527 device                 )

  $41 CAN_CONTROL  CAN!                 \ Set INIT state
  $41 CAN_CPUINT   CAN!                 \ SCLK = XTAL/2 (8Mhz)

  $40 CAN_BUSCONF  CAN!                 \ ???
  $00 CAN_BTR0     CAN!                 \ Set CANBUS speed to 500kbps
  $1C CAN_BTR1     CAN!
  $FF CAN_P2CONF   CAN!                 \ setup led pins for output

  $02 CAN_CONTROL  CAN!                 \ take 82527 out of config mode

  $FF CAN_GMS 0 + CAN!                 \ Global Mask = FFFF
  $FF CAN_GMS 1 + CAN!
  $FF CAN_GME 0 + CAN!                 \ Extended Mask = FFFFFFF8
  $FF CAN_GME 1 + CAN!
  $FF CAN_GME 2 + CAN!
  $F8 CAN_GME 3 + CAN!

  $02 CAN_CONTROL  CAN!                 \ take 82527 out of config mode

  CAN-FLUSH

  16 1 do
    defaultcan i ConfigureMessage
  loop

  $00 CAN_STATUS  CAN!                 \ Clear Status Reg
  can-flush
;
