\ TSEPOS 84 24 25 \
\ TSECFGEXPR    '^[ \t]*{\(\ \**}|{code}|{:}'
\ TSECFGOPTS    'ix'

\ T             title
\ P             begin paragraph (ends after any non * line)
\ H             a general heading
\ S             Section sub-title
\ D             Definition, next word is title, all following text
\               and * lines are descriptions
\ N             Forced newline
\ C             Description comment for Forth word
\ E             An Edit Comment line

\ =======================================================================
( **                                                                    )
( *T SMC9192/94 Ethernet Driver Code                                    )
( **                                                                    )
( *P The hardware driver layer for the Standard MicroSystems 9192/94    )
( ** Ethernet controller chips as implemented on the MPE H8 Board,      )
( **                                                                    )
( *P The controller has 64 * 8bit registers organised as four banks of  )
( ** 16.                                                                )
( **                                                                    )
\ =======================================================================

\ =======================================================================
( **                                                                    )
( *S Constants                                                          )
( **                                                                    )
( *D EtherBase          Hardware base address of Controller registers   )
( *D BankSelect         Offset of bankselect register                   )
( **                                                                    )
( *H Bank0 Registers                                                    )
( *D TCRL               Transmit Control Register - lo byte             )
( *D TCRH               Transmit Control Register - hi byte             )
( *D TXSTATUS           Transmit Status Register                        )
( *D RCR                Receive Control Register                        )
( *D MCR                Memory Control Register                         )
( **                                                                    )
( *H Bank1 Registers                                                    )
( *D Configl            Configuration Register - lo byte                )
( *D Configh            Configuration Register - hi byte                )
( *D IA0                Hardware Address, Ethernet MSB                  )
( *D IA1                Hardware addr                                   )
( *D IA2                Hardware addr                                   )
( *D IA3                Hardware addr                                   )
( *D IA4                Hardware addr                                   )
( *D IA5                Hardware addr                                   )
( *D Ctrh               Control Register - hi byte                      )
( **                                                                    )
( *H Bank2 Registers                                                    )
( *D MMU                Memory management unit cmd reg                  )
( *D PNR                Packet number register                          )
( *D ARR                Allocation result register.                     )
( *D Pointerl           Memory pointer register                         )
( *D Pointerh           Memory pointer register                         )
( *D Datal              Reg to send packets too...                      )
( *D Datah              Reg to send packets too...                      )
( *D IntStatus                                                          )
( *D IntMask                                                            )
( **                                                                    )
( *H Misc. Constants                                                    )
( *D RelRx              MMU Command to release memory from an rx        )
( *D RxRd               SMC Command to read received packet             )
( *D TxWr               SMC Command to write to tx area of ram          )
( *D AllocIntMask       Memory allocated mask                           )
( *D MaxMsgSize         Max 802.3 Ether frame size in bytes             )
( *D AllocTx                                                            )
( *D bit0               Bitmask                                         )
( *D bit1               Bitmask                                         )
( *D bit2               Bitmask                                         )
( *D bit3               Bitmask                                         )
( *D bit4               Bitmask                                         )
( *D bit5               Bitmask                                         )
( *D bit6               Bitmask                                         )
( *D bit7               Bitmask                                         )
( *D EvenTx             Control byte for tx of even bytes               )
( *D OddTx              Control byte for tx of odd bytes                )
( *D RxTask#            Multi Task ID                                   )
( **                                                                    )
\ =======================================================================

csext $8000 + constant EtherBase

$0E constant BankSelect
$00 constant Bank0
$01 constant Bank1
$02 constant Bank2
$03 constant Bank3

\ Bank 0 registers
$00 constant TCRL
$01 constant TCRH
$02 constant TxStatus
$05 constant RCR
$0A constant MCR

\ Bank 1 registers
$00 constant Configl
$01 constant Configh
$04 constant IA0
$05 constant IA1
$06 constant IA2
$07 constant IA3
$08 constant IA4
$09 constant IA5
$0D constant Ctrh

\ Bank 2 registers

$00 constant MMU
$02 constant PNR
$03 constant ARR
$06 constant Pointerl
$07 constant Pointerh
$08 constant Datal
$09 constant Datah
$0C constant IntStatus
$0D constant IntMask

\ Various bits and bobs...
$80 constant RelRx
$E0 constant RxRd
$40 constant TxWr
$08 constant AllocIntMask
#1530 constant MaxMsgSize
$20 constant AllocTx
$01 constant bit0
$02 constant bit1
$04 constant bit2
$08 constant bit3
$10 constant bit4
$20 constant bit5
$40 constant bit6
$80 constant bit7
$00 constant EvenTx
$20 constant OddTx
$01 constant RxTask#

\ =======================================================================
( **                                                                    )
( *S Hardware Interface Layer                                           )
( **                                                                    )
\ =======================================================================

: SetBank	\ bank -- ; bank = 0 to 3
( *G Select active register bank.                                       )
  $03 and
  EtherBase BankSelect + c!
;

: e!            \ val offset --
( *G Set a register contents to val.                                    )
  EtherBase + c!
;

: e@            \ offset -- val
( *G Read a register from selected bank.                                )
  EtherBase + c@
;

\ =======================================================================
( **                                                                    )
( *S Diagnostics                                                        )
( **                                                                    )
\ =======================================================================

: .bank         \ --
( *G Display register contents for currently selected SMC bank.         )
  $10 0 do
    i e@ .byte space loop cr
;

: .reg          \ --
( *G Display contents of all SMC Ethernet Controller's registers        )
  base @ >r hex
  cr ." Offset  " $10 0 do i .byte space loop cr cr
  #4 0
  do
    i setbank ." Bank " i . space .bank
  loop
  r> base !
;

\ =======================================================================
( **                                                                    )
( *S Driver Layer                                                       )
( **                                                                    )
\ =======================================================================

: AllocateMemory	\ n --
( *G Allocate n bytes of memory to hold a packet for transmission.      )
  bank2 setbank
  $10 +                         \ because we will add to the msg length later
  $100 /                        \ because we allocate in 256 byte chuncks
  AllocTx or                    \ add command and length together
  MMU e!                        \ send the rconstantest to ether

  begin
    IntStatus e@                \ get interupt status register
    AllocIntMask and            \ loop until bit set
    0=
  while
    pause
  repeat

  ARR e@ $1F and                \ extract allocated packet number
  PNR e!                        \ store at packet number register
;

: QueueTx       \ --
( *G Enqueue last written packet for transmission.                      )
  $C0 MMU e!                    \ enqueue packet command
;

: WritePacket   \ addr len --
( *G Write an Ethernet frame to chip for transmission                   )
  >r                            \ store len for later
  $00 Datah e!                  \ write status word
  $00 Datah e!
  r@ 6 +                        \ len+6,=status,length and control word
  $FFFE and                     \ ensure an even number
  $100 /mod swap                \ split len to two 8 bit
  Datah e! Datah e!             \ store low then high count byte
  r@ bounds                     \ addr len bounds for loop
  do
    i c@ Datah e!               \ get buffer entry, store to chip
  loop
  r> 2 mod                      \ was length odd ?
  if                            \ y:
    OddTx Datah e!              \ packet control byte
  else                          \ n:
    $00 Datah e!                \ dummy write
    EvenTx Datah e!             \ packet control byte
  endif
;

: init-SM	\ --
( *G Initialise SMC Chip                                                )
  cr ." Configure SMC Ether Device"
  cr ."   Reset MMU"
  Bank2 SetBank $40 MMU e!      \ Reset MMU
  cr ."   Interrupt Source (unused)"
  bank1 SetBank $00 Configl e!  \ Active HI ints
  cr ."   Twisted Pair media"
  Bank1 SetBank $00 Configh e!  \ select 10BaseT transmission media
  cr ."   Reserve TX frame space"
  Bank0 SetBank $06 MCR e!      \ reserves 6*256=1536 bytes for TX
;


: init-EtherTx  \ --
( *G Enable Ethernet TX module                                          )
  cr ." SMC Ether Device, Enable TX"
  Bank0 SetBank
  Tcrl e@ 1 or Tcrl e!
  Bank1 SetBank  $08 Ctrh e!    \ set auto release of transmit memory
;


: init-EtherRx  \ --
( *G Enable Ethernet RX module                                          )
  cr ." SMC Ether Device, Enable RX"
  Bank0 SetBank
  $03 Rcr e!                    \ enable receiver and strip rx crc
  Bank2 SetBank
  bit0 IntMask e!               \ enable rx ints on ether chip

;

: etheradd>CS   \ addr --
( *G Store 6 byte Ethernet address into SMC chip from memory buffer.    )
  cr ." SMC Ether Device, Hardware Address "
  Bank1 SetBank
  dup c@ dup .byte IA0 e! 1+ [char] - emit
  dup c@ dup .byte IA1 e! 1+ [char] - emit
  dup c@ dup .byte IA2 e! 1+ [char] - emit
  dup c@ dup .byte IA3 e! 1+ [char] - emit
  dup c@ dup .byte IA4 e! 1+ [char] - emit
      c@ dup .byte IA5 e!
;

: IsRx?         \ -- t|f
( *G Check if incoming data is present                                  )
  Bank2 SetBank
  IntStatus e@ 1 and
;

: get_ether_pkt { *dest maxlen | plen -- len }
( *G Get pending receive packet to a buffer                             )

  Bank2 SetBank
  $00 Pointerl e!               \ reset pointer value
  RxRd Pointerh e!              \ read rx ram area, auto inc low pntr
  Datah e@ drop                 \ discard hash code value
  Datah e@ drop                 \ high status contains odd/even info
  Datah e@ Datah e@ $100 * +    \ 2 * 8 bit -> 16 bit number
  $07FF and                     \ byte count only uses 11 bits !
  $05 -                         \ read 4 bytes out, ignore last control byte

  dup maxlen > if               \ ensure length < maxlen
    drop maxlen
  endif

  1 -                           \ HACK! SJC000
  dup -> plen

  0 ?do                         \ loop for rest of message
   datah e@                     \ read byte from ether chip ram
   *dest i + c!                 \ copy byte to buffer
  loop

  RelRx MMU e!                  \ remove and release memory from rx

  plen                          \ return packet length
;

: DiscardRX     \ --
( *G Throw away pending input packet due to lack of memory.             )
  Bank2 SetBank
  RelRx MMU e!                  \ remove and release memory from rx
;


: send_ether_pkt        \ addr len --
( *G Send packet from supplied buffer                                   )

  dup                           \ ensure packet is over 64 bytes or it will
  $40 < if                      \ be ignored by the SMC chipset
    drop $40
  endif

  Bank2 SetBank
  dup AllocateMemory            \ get memory for transmission
    $00 Pointerl e!             \ reset offset value
    TxWr Pointerh e!            \ set to write Tx ram area
    WritePacket                 \ send packet to ether chip
    QueueTx
;

: set-ether     \ e1 e2 e3 e4 e5 e6 --
( *G Set device 48 bit Ethernet address. )
  IA5 e!
  IA4 e!
  IA3 e!
  IA2 e!
  IA1 e!
  IA0 e!
;

: packet-listen	\ --
( *G Listens to the network and displays all the traffic. )
  init-cs8900				\ initialise chip
  1 2 3 4 5 6 set-ether			\ set a 48 bit Ethernet address
  init-EtherRx				\ now init the receiver
  init-EtherTx				\ & the transmitter
  begin
    isrx? if
      here 1536 get_ether_pkt
      here swap 1 max dump
    then
    key?
  until
;

