
\ ***************************************************************
\ Module:       d-cache.fth
\               Contains data cacheing functions
\ Author:       Jeremy Moller
\ Organisation: CEM Systems Ltd, Belfast
\ Copyright:    1999
\ ******************************************************************



struct CACHE-INFO
    int ->cache-link
    int ->mem-block
    int ->disk-block
    int ->cache-valid
    int ->cache-time
    int ->cache-dirty
end-struct

$125a5aff equ CACHE_VALID
$125a5a5a equ CACHE_DIRTY
$00000000 equ CACHE_CLEAN

#32 equ #CACHE_BLOCKS

#CACHE_BLOCKS CACHE-INFO * equ CACHE_HDR_SIZE
#CACHE_BLOCKS BLOCK_SIZE * equ CACHE_SIZE

CACHE_SIZE      buffer: cache-memory        \ the cache blocks
CACHE_HDR_SIZE  buffer: cache-hdr-table     \ the cache lookup table

variable active-c-block
variable free-c-blocks


\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: init-cache { | l_cache -- }
    cr ." Init Disk cache     - Done " 
    cache-sema signal
    cache-hdr-table CACHE_HDR_SIZE erase    \ clear file control block table
    cache-memory CACHE_SIZE erase
    0 active-c-block !
    0
    #CACHE_BLOCKS 0 do
        i CACHE-INFO * cache-hdr-table + -> l_cache
        i BLOCK_SIZE * cache-memory + l_cache ->mem-block !
        l_cache ->cache-link !
        l_cache ->cache-link
    loop
    free-c-blocks !
    ;
    

: (get-cache)     \ -- struct ;
    free-c-blocks @ dup if
        dup ->cache-link @ free-c-blocks !  	\ unhook from free chain
        active-c-block @ over ->cache-link !    \ add to active chain
        dup active-c-block !
    endif ;

: release-cache \ struct -- ; exit quietly if FCB can't be found
    active-c-block
    begin                               \ -- struct s2
        2dup ->cache-link @ <>          \  MUST be first
    while
        ->cache-link @
        dup 0= if                       \ if at end of chain
            2drop  exit                 \   get out quietly
        endif   
    repeat
\ struct prev -- ; take the next link, and place it in the previous link
    over ->cache-valid off
    over ->cache-link @  swap ->cache-link !
    free-c-blocks @ over ->cache-link ! \ add to free chain
    free-c-blocks !
;

\ -----------------------------------------------------------------
\ ENTRY, a possible oldest cache structure
: get-cache ( ?struct -- struct )
    free-c-blocks @ if          	\ any free cache blocks ?
        drop                    	\ YES don't need old cache
        (get-cache)             	\ try to get fresh cache
    else                        	\ NO
        dup get-oldest-cache    	\ oldest cache
    endif ;
    
: get-oldest-cache { l_struct | -- }
    l_struct ->cache-dirty @ CACHE_DIRTY = if
        l_struct ->disk-block @
        l_struct ->mem-block @ write-block drop
    endif
    l_struct ->disk-block off ;
    
: blk-in-cache? { l_blk | l_struct l_oldest l_found -- struct t/f )
    -1 -> l_oldest
    0 -> l_struct
    false -> l_found
    active-c-block
    begin
        @ ?dup                  	\ still in chain
    while
        dup ->disk-block @ l_blk = if   \ this block
            true exit
        endif
        dup ->cache-time @ l_oldest u< if
            dup -> l_struct
            l_struct ->cache-time @ -> l_oldest
        endif           
    repeat
    l_struct l_found ;


\ ------------------------------------------------------------------

: (read-blk) ( blk -- struct )
    dup blk-in-cache? if
        nip                 		\ YES, lose block, keep structure
    else
        read-blk>c          		\ NO, read into cache
    endif ;
    

: read-blk>c { l_blk l_struct | l_cache -- l_cache }
    l_struct get-cache dup if           	\ able to get cache ?
        -> l_cache
        CACHE_CLEAN l_cache ->cache-dirty ! 	\ signal it's clean
        CACHE_VALID l_cache ->cache-valid ! 	\ signal it's valid
        ticks l_cache ->cache-time !  		\ record the time
        l_blk l_cache ->disk-block !  		\ record the block
        l_blk l_cache ->mem-block @ read-f-block
        l_cache and
    endif ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\ a partial block read
: read-p-blk { l_lba l_buff l_start l_size | -- t/f }
    cache-sema request
    l_lba (read-blk)            \ read and return cache structure
    ticks over ->cache-time !   \ record the time
    ->mem-block @ l_start +     \ source
    l_buff                      \ destination
    l_size                      \ amount
    cmove
    true 
    cache-sema signal
    ;

\ a full block read
: read-f-blk { l_lba l_buff | -- t/f }
    cache-sema request
    l_lba blk-in-cache? if                  	\ this block's in cache
        ->mem-block @ l_buff BLOCK_SIZE cmove   \ copy from cache
        cache-sema signal
        true
    else
        cache-sema signal
        drop
        l_lba l_buff read-f-block               \ no, so just read full block
    endif
    ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: write-p-blk { l_blk l_buffer l_start l_end | l_struct -- t/f  }
    cache-sema request
    l_blk (read-blk) dup if
        -> l_struct     			\ find a cache block
        l_buffer                            	\ source
        l_struct ->mem-block @  l_start +   	\ destination
        l_end l_start -                     	\ amount
        cmove                               	\ write to cache
        ticks l_struct ->cache-time !       	\ record time in cache
        CACHE_DIRTY l_struct ->cache-dirty !    \ mark as dirty
        true
    endif
    cache-sema signal
    ;


\ does a full block write to disk, replaces WRITE-BLOCK
: write-f-blk { l_lba l_buff | -- t/f }
    cache-sema request
    l_lba blk-in-cache? if          	\ is there a partial in cache ?
        release-cache               	\ YES, well release it
    else
        drop
    endif
    cache-sema signal
    l_lba l_buff write-block        	\ write full block to disk
    ;
    
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: release-all-cache ( -- )
    di
    active-c-block ->cache-link @
    begin
         ?dup
    while
        dup ->cache-link @      	\ get next link
        swap release-cache
    repeat
    ei ;

: sync_vol { | l_struct -- }
    cache-sema request
    active-c-block ->cache-link @
    begin
         ?dup
    while
        -> l_struct
        l_struct ->cache-valid @ CACHE_VALID =
        l_struct ->cache-dirty @ CACHE_DIRTY = and if
            l_struct ->disk-block @
            l_struct ->mem-block @ write-block drop
            l_struct ->cache-dirty off
            ticks l_struct ->cache-time ! \ record time in cache
        endif
        l_struct ->cache-link @     	\ get next link
    repeat
    cache-sema signal ;

#15000 equ VOL_TIMEOUT

: time_sync_vol { | l_struct l_expired -- }
    ticks VOL_TIMEOUT - -> l_expired
    cache-sema request
    active-c-block ->cache-link @
    begin
         ?dup
    while
        -> l_struct
        l_struct ->cache-valid @ CACHE_VALID =
        l_struct ->cache-dirty @ CACHE_DIRTY = and 
        l_struct ->cache-time @ l_expired u< and if
            l_struct ->disk-block @ l_struct ->mem-block @ write-block drop
            l_struct ->cache-dirty off          \ mark as clean
            ticks l_struct ->cache-time !       \ record time in cache
        endif
        l_struct ->cache-link @     		\ get next link
    repeat
    cache-sema signal ;


\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ display list of active cache blocks
: show-cache ( -- )
    active-c-block (show-cache) ;

\ show list of free cache blocks
: show-f-cache ( -- )
    free-c-blocks (show-cache) ;

\ --------------------------------------------------------------

: (show-cache) { | l_struct -- }
    begin
        ->cache-link @ ?dup
    while
        -> l_struct
        cr
        l_struct .hex
        l_struct ->cache-valid @ CACHE_VALID =
        if ." V " else ." N " endif
        l_struct ->cache-dirty @ CACHE_DIRTY =
        if ." D " else ." C " endif
        l_struct ->mem-block @ .hex space
        l_struct ->disk-block @ .decimal space
        l_struct ->cache-time @ .decimal space
        l_struct
    repeat ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: t-c1
    #30 0 do
        i (read-blk) cr .hex
    loop ;

