
\ ***************************************************************
\ Module:       fcb.fth
\               Contains file control block functions
\ Author:       Jeremy Moller
\ Organisation: CEM Systems Ltd, Belfast
\ Copyright:    1999
\ ******************************************************************


#15 equ SYS_NFCB        \ max number of different concurrently open files
#30 equ TASK_NCFILE     \ max number of concurrent open files

SYS_NFCB FLD-REC *  equ FCB-SIZE
TASK_NCFILE open-table-struct * equ OPENT-SIZE

FCB-SIZE    buffer: FCB-TABLE
OPENT-SIZE  buffer: OPEN-TABLE

FLD-REC cell - equ FCB_DATA_REC \ the 1st element in struct must be preserved

FLD-REC buffer: cdir-buffer

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ returns disk block of required FLD
: fld-d-block ( fld -- blk )
    FLD/BLOCK / FLD0_BLOCK + ;

\ returns block offest of required FLD
: fld-m-block ( fld -- offset )
    FLD/BLOCK mod FLD-REC * ;

\ -------------------------------------------------------------------

\ update specific FLD to disk via cache
: rewrite-fld { l_fld l_buff | -- t/f }
    l_fld fld-d-block       \ disk block
    l_buff                  \ destination
    l_fld fld-m-block       \ start
    dup FLD-REC +           \ end
    write-p-blk ;

\ read in specific FLD from disk via cache
: read-in-fld { l_fld l_buff | -- t/f }
    l_fld fld-d-block       \ disk block
    l_buff cell+            \ destination
    l_fld fld-m-block cell+ \ start
    FCB_DATA_REC            \ size
    read-p-blk ;

\ ---------------------------------------------------------------

\ RETURN file control block number if successful else -1
: create-fcb { l_fld | l_fcb -- FCB }
    l_fld existing-fcb? ?dup 0= if  \ already there, return struct
        get-fcb dup if          \ 1. find free FCB
            -> l_fcb
            l_fld l_fcb read-in-fld drop
            l_fld       l_fcb ->fld-no !        \ FLD in the fcb
                        l_fcb start-fcb-open    \ number of tasks using this file
            l_fcb
        endif
    else
    \   ." Already open "
        dup inc-fcb-open
    endif ;

((

\ the FID is now used for all read and write operations
: create-fot ( FCB -- FID )
    \ 1. find free open file table
    \ 2. put FCB number in File Open Table
    \ 3. clear read write pointer
    \ 3. increment "->open-tables" in FCB
    \ RETURN File ID number if successful.
    ;

))

#8192                                   \ timer errors
  dup constant err_no_fcbs  1+
  dup constant err_bad_fcbs  1+
  dup constant err_no_otbs  1+
  dup constant err_bad_otbs  1+
drop

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

variable active-fcbs
variable free-fcbs
variable active-otbs
variable free-otbs

\ file control blocks

: init-fcbs ( -- )
    fcb-table FCB-SIZE erase        \ clear file control block table
    0 active-fcbs !
    0
    fcb-table FCB-SIZE bounds do
        i ->next-fcb !
        i
    FLD-REC +loop
    free-fcbs !
    ;
    

: get-fcb     \ -- struct ;
    free-fcbs @ dup if
        di                                    \ make this safe!
        dup @ free-fcbs !                   \ unhook from free chain
        active-fcbs @ over ->next-fcb !         \ add to active timer chain
        dup active-fcbs !
        ei
    endif ;

: release-fcb \ struct -- ; exit quietly if FCB can't be found
  active-fcbs
  begin                                 \ -- struct s2
    2dup ->next-fcb @ <>                \ ->next-fcb MUST be first
   while
     ->next-fcb @
     dup 0=                             \ if at end of chain
     if  2drop  exit  endif             \   get out quietly
  repeat
\ struct prev -- ; take the next link, and place it in the previous link
  over ->next-fcb @  swap ->next-fcb !
  free-fcbs @ over ->next-fcb !          \ add to free chain
  free-fcbs !
;

: existing-fcb? \ fld -- struct ; 
    active-fcbs                 	\ test active chain
    begin                       	\ -- fld s2
        2dup ->fld-no @ <>      	\ see if file already open 
    while
        ->next-fcb @
        dup 0= if                       \ if at end of chain
            nip exit
        endif             		\   get out quietly
    repeat
    nip ;                   		\ lose fld

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: set-fld-in-fcb ( fld fcb -- )
    ->fld-no !
    ;

: inc-fcb-open ( fcb -- )
    ->open-tables 1 swap +! ;

: start-fcb-open ( fcb -- )
    ->open-tables 1 swap ! ;

\ returns true if all access closed
: dec-fcb-open ( fcb -- t/f )
    ->open-tables dup @ dup if
        1- tuck swap !
        0=                  		\ all closed ?
    else
        2drop
        true                		\ already closed
    endif
 ;

: file-open? ( *name len -- t/f )
    get_fn dup if
        fn_open?
    endif ;

\ use the Fn to find if named file is open
: fn_open? ( fn -- t/f )
    existing-fcb? 0<> ;     \ find out if there's already a file control block.

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\ open table blocks, used by task.

: init-otbs ( -- )
    OPEN-TABLE OPENT-SIZE erase     	\ clear file open table
    0 active-otbs !
    0
    OPEN-TABLE OPENT-SIZE bounds do
        i ->next-otb !
        i ->next-otb
    OPEN-TABLE-STRUCT +loop
    free-otbs !
    ;

: create-otb     \ -- struct ; 
    free-otbs @ dup if
        di                              \ make this safe!
        dup ->next-otb @ free-otbs !    \ unhook from free chain
        active-otbs @ over ->next-otb ! \ add to active timer chain
        VALID_OTB over ->valid-otb !
        dup active-otbs !
        ei
    endif
;

: release-otb \ fid -- ; 
    active-otbs
    begin                               \ -- struct s2
        2dup ->next-otb @ <>            \  MUST be first
    while
        ->next-otb @
        dup 0= if                       \ if at end of chain
            2drop
             exit                	\   get out quietly
        endif          
    repeat
\ struct prev -- ; take the next link, and place it in the previous link
    over ->fcb-number off
    over ->valid-otb off
    over ->next-otb @  swap ->next-otb !
    free-otbs @ over ->next-otb !       \ add to free chain
    free-otbs !
     ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: r/w-ptr-inrange? ( fid -- t/f )
    dup get-r/w-b-ptr
    swap fcb-addr ->file-blocks @ <= ;

: r-ptr-inrange? ( fid -- t/f )
    dup get-f-r/w-ptr
    swap fid-file-size <= ;

: fid-file-size ( fid -- size )
    fcb-addr ->file-size @ ;

: set-f-r/w-ptr ( value fid -- )
    ->r/w-data-ptr ! ;

: get-f-r/w-ptr ( fid -- value )
    dup if
        ->r/w-data-ptr @
    endif ;

: get-r/w-b-ptr ( fid -- block )
    dup if
    ->r/w-data-ptr @ BLOCK_SIZE /
    endif ;


: get-r/w-pb-ptr ( fid -- offset )
    dup if
        ->r/w-data-ptr @ BLOCK_SIZE mod
    endif ;

: next-f-r/w-ptr ( fid -- )
    dup if
    ->r/w-data-ptr BLOCK_SIZE swap +!
    endif ;
    
: inc-f-r/w-ptr ( step fid -- )
    ->r/w-data-ptr tuck @ + swap ! ;
    
: r/w-block { l_fid | l_hd-block l_block l_fcb -- phys-blk } 
    -1 -> l_hd-block
    l_fid get-r/w-b-ptr -> l_block  
    l_fid fcb-addr -> l_fcb
    l_fcb ->file-blocks @ l_block > if  \ test if pointer inside allocated blocks
        l_fcb ->extent-start-0 EXTENT_LIST bounds do
            i ->extent-size @ ?dup if   \ get extent size
                l_block swap - dup 0< if \ found extent
                    drop        
                    i ->extent-start @ l_block + \ get extent start 
                    -> l_hd-block
                    leave
                else
                    -> l_block          \ rem-blocks
                endif
            else
                leave
            endif
        FILE-EXTENTS +loop
        l_fcb ->indirect-block @ 0<>    \ indirect extent defined
        l_hd-block -1 = and if          \ not found so far
            l_fid l_block get-indirect-extent -> l_hd-block
        endif
    else
        -2 -> l_hd-block        	\ need more blocks allocated
    endif
    l_hd-block ;

\ returns the FCB address from the FID
: fcb-addr ( fid -- fcb )
    ->fcb-number @  ;
    
\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

