
\ ***************************************************************
\ Module:	file-bitmap.fth
\		Contains file block allocation functions
\ Author:	Jeremy Moller
\ Organisation:	CEM Systems Ltd, Belfast
\ Copyright:	1999
\ ******************************************************************

((
These functions are used for indicating disk usage.
A space is allocate were one bit represents one block (512 bytes) on
the media.

When a block is in use, the appropiate bit in the BITMAP table will
be set to one.

For groups of blocks greater than 31, these will always start on a
multiple of 32 blocks.
))

variable p-blocks	\ groups of less 32blocks required
variable no-32		\ number of 32block groups required
variable b-ptr		\ disk buffer search pointer
variable grab/free	\ true for grab, false for release

BLOCKS/BIT_TABLE equ BLOCKS/GROUP
$FFFFFFFF equ 32BLOCKS

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\  Need-blocks: This attempts to find a contigous area of required space
\				and allocates it.
\ ENTRY: Number of 512byte blocks required
\ EXIT:	-1 can't allocate
\ EXIT:	Start block number on media

: need-blocks ( amount -- start )
  dup find-free-blocks dup true <> if
    \ cr ." Found start block @ " dup .
    tuck swap				\ keep start pointer
    grab/free off			\ flag to grab blocks
    block-order allocate-blocks
  else
    cr ." Error finding space "
    nip
  endif ;

\ free-blocks:	This will release the required number of blocks from teh start point
: free-blocks ( start-block amount -- )
  grab/free on				\ flag to release blocks
  block-order allocate-blocks ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

: find-free-blocks { l_req-blocks | l_start-block l_32-found -- start-block }
  l_req-blocks #32 /mod no-32 !		\ number of complete 32 blocks
  p-blocks !				\ number od partial 32 blocks
  0 -> l_32-found			\ number of 32 blocks found
  -1 -> l_start-block
  USAGE_BLOCK0 USAGE_TABLE_SIZE bounds ?do
    i
    \ cr ." Read " dup u.
    bit_block_buffer read-hd-block drop	\ read in block
    b-ptr off				\ start block pointer
    begin
      no-32 @ l_32-found - 0<> if	\ complete 32 blocks to find
	bit_block_buffer b-ptr @ + @ 0= if	\ area free
	  l_start-block -1 = if		\ first of the free blocks
	    \ cr ." Found 1st " 
	    b-ptr @ 08 *
	    i USAGE_BLOCK0 - BLOCKS/BIT_TABLE * +
	    \ dup .
	    -> l_start-block		\ record it
	  endif
	  l_32-found 1+ -> l_32-found	\ another block found
	else
	  -1 -> l_start-block		\ cancel find and restart area search
	  0 -> l_32-found		\ 
	endif
      else
	find-<32blocks dup -1 = if	\ reset and continue search
	  -> l_start-block		
	  0 -> l_32-found		\ 
	else				\ all done
	  no-32 @ if			\ multiple blocks
	    \ ." Continue block free, done "
	    drop
	  else				\ single block 
            \  ." Partial block free, done " dup .
	    b-ptr @ 08 * +
	    i USAGE_BLOCK0 - BLOCKS/BIT_TABLE * + -> l_start-block
	  endif
	  -1 -> l_req-blocks
	endif
      endif
      b-ptr @ cell+ dup b-ptr ! BLOCK_SIZE >=	\ these blocks searched
      l_req-blocks -1 = or		\ OR all required found
    until
    l_req-blocks -1 = if
      leave
    endif
  loop
  l_req-blocks -1 = if			\ Yep, find required
    l_start-block
  else
    -1											\ Nop, could not find
  endif ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

\ no find of continue 		=> -1 resets "start block" pointer, continue search
\ no find of small block	=> -1 continue search
\ find of continue		=>  0 all found and done
\ find of small block		=> ptr all found done, set "start block" pointer

: find-<32blocks ( -- result )		\ true is NOT FOUND else start block
  bit_block_buffer b-ptr @ + @ dup -1 <> if	\ read 32 blocks at pointer
\ ." some space "
    p-blocks @					\ number of disk blocks required
    no-32 @ if					\ continuation of blocks
      reverse-bitmap 
      and 0<> 					\ test required area free
    else					\ small block on own
      \ ." on own "
      (find-partial-block) 
    endif
  endif ;

\ this function creates a bit pattern starting from the highest bit
\   1 bit is $80000000
\  4 bits is $f0000000
\ 16 bits is $ffff0000

: reverse-bitmap ( size -- pattern )
  1 over lshift 1- 			
  swap #32 swap - lshift ;

\ -----------------------------------------------------


\ finds partial block of SIZE anywhere in 32 blocks
\ returns -1 if it won't fit.
: (find-partial-block) { l_data l_size | l_block-no -- start-block-no }
  -1 -> l_block-no
  l_size #32 ?do
    1 l_size lshift 1- i l_size - lshift
    l_data and 0= if
      #32 i - -> l_block-no
      leave
    endif
  -1 +loop
  l_block-no ;

\ -------------------------------------------------------------

: block-order { l_start l_amount | l_buff -- buffer }
  block-alloc allocate if
    drop  false exit
  endif
  -> l_buff
  l_start BLOCKS/GROUP /mod			\ start position block number
  USAGE_BLOCK0 + l_buff ->start-block !
  dup l_buff ->start-address !
  BLOCKS/GROUP swap - dup l_buff ->first-amount ! \ space remaining
  l_amount swap -				\ additional amount of space required
  dup 0< if					\ it'll fit one page
    drop 0
    l_amount l_buff ->first-amount !
  endif
  BLOCKS/GROUP /mod
  l_buff ->#blocks !
  l_buff ->last-amount !
  l_buff ;

\ ENTRY structure
: allocate-blocks { l_buff | } 
  l_buff ->start-block @
  l_buff ->start-address @
  l_buff ->first-amount @ grab-first-block
	
  l_buff ->start-block @ 1+
  l_buff ->#blocks @ bounds ?do
     \ cr ." full block no.  " i .
     i grab-32complete-blocks
  loop
	
  l_buff ->last-amount @ ?dup if
    l_buff ->start-block @ 1+
    l_buff ->#blocks @ + grab-last-block
  endif
  l_buff free drop
;

\ ---------------------------------------------------------------
: take/free ( data1 data2 -- 'data )
  grab/free @ if			\ true for release
    not and
  else					\ false for take
    or
  endif ;

: take/free32 ( -- 'data )
  grab/free @ if			\ true for release
    0
  else					\ false for take
    32BLOCKS
  endif ;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
	
: grab-first-block ( start-block start-address first-amount -- )
  2 pick				\ get start block
\ cr ." first U block " dup .
  bit_block_buffer read-hd-block drop	\ read in block
\ ." Marking  "
  over adjust-blk-addr bit_block_buffer + \ start address of 32 bit units
  over adjust-blk-addr 			\ last address of 32 bits units
  bounds
\ 2dup ." From " .lword ."  To " .lword
  ?do
    take/free32 i !
  cell +loop
					\ s-addr f-amount
  dup #32 mod ?dup if			\ need a partial block
  \ ."  And " 				\ s-addr fa famod
    reverse-bitmap	
    2 pick #32 mod rshift		\ data to write
    -rot 				\ b-map s-addr f-amount 
    adjust-blk-addr swap adjust-blk-addr +	\ address to write
    bit_block_buffer +
  \ 2dup swap .lword ."  @ " .lword
    dup @ rot take/free swap !
  else					\ don't need partial block
    2drop
  endif
  bit_block_buffer write-hd-block drop	\ write out block
;

: adjust-blk-addr ( addr1 -- addr2 )
  #32 / cells ;

: grab-32complete-blocks ( block -- ) 
  dup
  bit_block_buffer read-hd-block drop	\ read in block
  bit_block_buffer BLOCK_SIZE bounds do
    take/free32 i !
  cell +loop
  bit_block_buffer write-hd-block drop	\ write back block
;

: grab-last-block ( last-amount last-block -- )
\ cr ." Last block " dup u.
  tuck
  bit_block_buffer read-hd-block drop	\ read in block
  dup #32 / cells dup
\ ." To " dup .
  bit_block_buffer swap bounds ?do
    take/free32 i !
  cell +loop

  swap #32 mod
  reverse-bitmap
  swap bit_block_buffer + 
\ 2dup swap .lword ."  @" .lword
  dup @ rot take/free swap !
\ cr ." Last block " dup u.
  bit_block_buffer write-hd-block drop	\ write back block
;

\ \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
\ End of file

: d-fbm { | l_*buff -- }
  BLOCK_SIZE allocate if
    ." Alloc error " exit
  endif
  -> l_*buff
  USAGE_BLOCK0 USAGE_TABLE_SIZE bounds ?do
    i l_*buff read-hd-block if		\ read in block
      cr ." Block" i .
      l_*buff BLOCK_SIZE dump
      cr ." Prees key " key upc ascii D = if
	unloop l_*buff free drop  exit
      endif
    else
      cr ." Disk error"
      unloop l_*buff free drop  exit
    endif
  loop
  l_*buff free drop ;



