\ AT24C128 Serial EEPROM driver

((
Copyright (c) 2002, 2004
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
))

\ *! at24c128
\ *T AT24C128 I2C driver
\ ===============
\ *S Introduction
\ ===============
\ *P The code for this 16k*8 I2C EEPROM was tested on an
\ ** MPE ARM USB Stamp board.


\ ====================
\ *S 24C128 primitives
\ ====================

$0A0 constant 24c128_id0	\ -- $A0 ; up to four devices
\ *G The base device address (devaddr) of a 24C128.

24c128_id0 equ EEbase	\ -- devaddr
\ *G Device address of the first EEPROM.

$04000 constant 24c128_size	\ -- 16k
\ *G Memory size of a 24C128 device.

$040 constant 24c128_page	\ -- 128
\ *G Size of a 24C128 page to write.

#10 constant 24c128_twr		\ -- n ;
\ *G The write time in milliseconds at 2.7v or above. Note
\ ** that at 1.8v, this value must be increased to 20 ms.

: 24c128_send_addr	\ Eaddr devaddr --
\ *G Issue a device address and EEPROM address.
  write_byte_i2c			\ device address
  dup 8 rshift write_byte_i2c		\ addr hi
  $FF and write_byte_i2c		\ addr lo
;

: 24c128_write_byte	\ byte Eaddr devaddr --
\ *G Write byte to the EEPROM devaddr at EEPROM address Eaddr.
\ ** Note that the system must wait at least 24C128_TWR milliseconds
\ ** before accessing the same device again.
  +ack_i2c  start_i2c
  $0FE and 24c128_send_addr		\ devaddr and addr
  write_byte_i2c			\ byte
  stop_i2c
;

: 24c128_write_page	\ src len Eaddr devaddr --
\ *G Write len bytes of memory at src to the EEPROM devaddr
\ ** starting at EEPROM address Eaddr. If len is 0, the address
\ ** Eaddr is written to the EEPROM and becomes the current
\ ** address, although no data is written.
\ ** Note that the system must wait at least 24C128_TWR milliseconds
\ ** before accessing the same device again.
  +ack_i2c  start_i2c
  $0FE and 24c128_send_addr		\ devaddr and addr
  bounds ?do
    i c@ write_byte_i2c
  loop
  stop_i2c
;

: 24c128_read_byte	\ Eaddr devaddr -- byte
\ *G Using EEPROM devaddr, read the byte at EEPROM address Eaddr.
  +ack_i2c  start_i2c
  tuck $0FE and 24c128_send_addr	\ devaddr and addr by dummy write
  start_i2c
  $001 or write_byte_i2c		\ devaddr again
  -ack_i2c  read_byte_i2c		\ read byte
  stop_i2c
;

: 24c128_read_block	\ dest len Eaddr devaddr --
\ *G Using EEPROM devaddr, read len bytes at EEPROM address Eaddr
\ ** to the buffer at dest.
  2 pick 0= if  2drop 2drop  exit  endif
  +ack_i2c  start_i2c
  tuck $0FE and 24c128_send_addr	\ devaddr and addr by dummy write
  start_i2c
  $001 or write_byte_i2c		\ devaddr again
  1-  2dup bounds			\ for all but the last byte
  ?do  read_byte_i2c i c!  loop
  -ack_i2c  read_byte_i2c -rot + c!	\ the last byte
  +ack_i2c  stop_i2c
;

: 24c128_read_seq	\ dest len devaddr --
\ *G Read len bytes to dest from the current/next EEPROM address.
  over 0= if  drop 2drop  exit  endif
  +ack_i2c  start_i2c
  $001 or write_byte_i2c		\ devaddr
  1-  2dup bounds			\ for all but the last byte
  ?do  read_byte_i2c i c!  loop
  -ack_i2c  read_byte_i2c -rot + c!	\ the last byte
  +ack_i2c  stop_i2c
;


\ *******************
\ *S Simple utilities
\ *******************
\ *P These are simple words to read and write data into the serial
\ ** EEPROM. They are not optimised for speed, but they do permit
\ ** easy use. Much faster operation may be achieved using SEPROM@
\ ** and SEPROM! below.

: eeb@		\ eaddr -- b
\ *G Read a byte (8 bits) from address EADDR in the chip.
  EEbase 24c128_read_byte
;

: eew@		\ eaddr -- w
\ *G Read a word (16 bits) from address EADDR in the chip.
  dup eeb@  swap 1+ eeb@		\ -- b0 b1
  8 lshift or				\ -- w
;

: eel@		\ eaddr -- x
\ *G Read a long (32 bits) from address EADDR in the chip.
  dup eew@  swap 2+ eew@		\ -- w0 w1
  #16 lshift or				\ -- x
;

: eeb!		\ b eaddr --
\ *G Write a byte (8 bits) to address EADDR in the chip.
  EEbase 24c128_write_byte  24c128_twr ms
;

: eew!		\ w eaddr --
\ *G Write a word (16 bits) to address EADDR in the chip.
  2dup eeb!				\ low byte
  swap 8 rshift swap 1+ eeb!		\ high byte
;

: eel!		\ x eaddr --
\ *G Write a long (32 bits) to address EADDR in the chip.
  2dup eew!				\ low word
  swap #16 rshift swap 2+ eew!		\ high word
;


\ ==================
\ *S EDSL primitives
\ ==================

: NextBinAlign	\ addr n -- addr'
\ *G Align an address to the next n-byte boundary, where n must
\ ** be a power of 2.
  tuck +  swap negate and		\ n 1- +  n negate and
;
  
: 24c128_read_len	\ len seprom -- len'
\ *G Determine how much can be read in one go.
  dup 24c128_size NextBinAlign		\ -- len seprom seprom'
  swap - min				\ -- len'
;

: 24c128_write_len	\ len seprom -- len'
\ *G Determine how much can be written in one go.
  dup 24c128_page NextBinAlign		\ -- len seprom seprom'
  swap - min				\ -- len'
;


\ =================
\ *S EDSL functions
\ =================

: SEPROM!	\ sram-addr length(bytes) seprom-addr --
\ *G Write length bytes to EEPROM address seprom-addr
\ ** from the buffer at sram-addr.
  over 0= if  drop drop drop  exit  endif
  $3FFFF and				\ max 256k in four devices
  24c128_id0 #16 lshift or		\ or in base devaddr
  begin					\ -- sram-addr len EDaddr
    2dup 24c128_write_len >r		\ -- sram-addr len EDaddr ; R: -- len'
    2 pick r@ 2 pick			\ -- s l e s l' e ; R: -- len'
    dup 24c128_size 1- and		\ offset into EEPROM, then get devaddr
    swap #16 rshift			\ -- s l e s l' e d ; R: -- len'
    24c128_write_page			\ -- s l e ; R: -- len'
    24c128_twr ms			\ wait for the write to complete
    rot r@ + -rot			\ update sram-addr
    swap r@ - swap			\ update len
    r> +				\ update EDaddr
    over 0=
  until					\ -- sram-addr len EDaddr
  drop drop drop
;

: SEPROM@	\ sram-addr length(bytes) seprom-addr --
\ *G Read length bytes at EEPROM address seprom-addr
\ ** to the buffer at sram-addr.
  over 0= if  drop drop drop  exit  endif
  $3FFFF and				\ max 256k in four devices
  24c128_id0 #16 lshift or		\ or in base devaddr
  begin					\ -- sram-addr len EDaddr
    2dup 24c128_read_len >r		\ -- sram-addr len EDaddr ; R: -- len'
    2 pick r@ 2 pick			\ -- s l e s l' e ; R: -- len'
    dup 24c128_size 1- and		\ offset into EEPROM, then get devaddr
    swap #16 rshift			\ -- s l e s l' e d ; R: -- len'
    24c128_read_block			\ -- s l e ; R: -- len'
    rot r@ + -rot			\ update sram-addr
    swap r@ - swap			\ update len
    r> +				\ update EDaddr
    over 0=
  until					\ -- sram-addr len EDaddr
  drop drop drop
;


\ *********
\ test code
\ *********

TestI2C? [if]

$400 buffer: xbuff

create stringbuff
  ", It's a major shennanigan, this one!"

: rb  $A0 24c128_read_byte  ;
: wb  $A0 24c128_write_byte  ;
: rbl $A0 24c128_read_block  ;
: wp  $A0 24c128_write_page  ;
: rs  $A0 24c128_read_seq  ;
[then]


\ ======
\ *> ###
\ ======

