\ UnixTime.fth - Unix style time based on LSECONDS ticker.

((
Copyright (c) 2003
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
))


\ ===========
\ *! unixtime
\ *T Unix-style time handling
\ ===========
\ *P Unix time handling is based on a 32 bit seconds counter,
\ ** in this case held as a variable called LSECONDS. All
\ ** calculations derive from LSECONDS. Second 0 is midnight
\ ** at the start of 1 Mar 1995. This will roll over on xx Mon yyyy.
\ ** If longer timekeeping is required, extend the system to use
\ ** 64 bit XSECONDS where LSECONDS contains the low 32 bits.
\ ** It is the user's responsibility to update LSECONDS.

\ *P Since many RTCs are attached by I2C or similar serial
\ ** connections, we suggest that the hardware either use
\ ** a one second interrupt from the RTC after the initial read,
\ ** or periodically resynchronise the ticker interrupt and derive
\ ** LSECONDS from the ticker interrupt.


\ ****************
\ *S Date handling
\ ****************
\ *P These routines are taken from the Julian calendar in
\ ** EXAMPLES\CALENDAR.FTH. They handle dates by converting
\ ** then to a single precision Julian day number. Day 0 is
\ ** 1 March 1900.

: date>         \ day month year -- jday ; see month codes
\ Given day month and year generate the day number. 
  dup 99 > if 1900 - then >r            \ year<100: XXth century
  dup 3 < if 12 + r> 1- >r then    	\ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  r@ 365 um* drop + r> 4 / +            \ and add 3/1/y - 3/1/0
;

: weekday       \ jday -- n ; 0=Sunday
\ *G Given a Julian day return the day of week number where
\ ** 0 is Sunday, 1 is Monday and so on.
  4 + 0 7 um/mod drop
;

: >date         \ jday -- day month year
\ *G Convert a Julian day number to day month year format.
  0 1461 um/mod                         \ no leap-year periods
  4 * >r                                \ convert to no years
  365 /mod                              \ no years and days
  dup 4 = if 2drop 365 3 then           \ correct if Febr.29th
  r> + >r                               \ -- days | years
  123 + dup 5 153 */                    \ compute month
  dup >r 153 5 */ - r>                  \ compute day
  1- dup 12 > if 12 - r> 1+ >r then     \ correct month and year
  r> 1900 +                        	\ -- day month year
;


\ ************************
\ *S Time/Date conversions
\ ************************

#24 #60 * #60 * equ secs/day	\ -- 86400
\ *G Seconds per day.
#60 #60 * equ secs/hr		\ -- 3600
\ *G Seconds per hour
#60 equ secs/min		\ -- 60
\ *G Seconds per minute
#60 equ mins/hr			\ -- 60
\ *G Minutes per hour
#24 equ hrs/day			\ -- 24
\ *G Hours per day
#34698 equ uday0		\ -- 34698
\ *G The Julian day number for Unix day 0 at 1 Mar 1995.

compiler
: j>uday	\ jday -- uday
\ *G Convert a Julian day based at 0 for 1 March 1900 to a Unix
\ ** day number derived from LSECONDS.
  uday0 -
;

: u>jday	\ jday -- uday
\ *G Convert a Unix day number as used by LSECONDS to a Julian
\ ** day based at 0 for 1 March 1900.
  uday0 +
;
target

[undefined] /umod [if]
: /umod		\ u1 u2 -- urem uquot
\ *G Single precision division of unsigned U1 by U2 returning
\ ** both remainder and quotient.
  0 swap um/mod
;
[then]

: time&date	\ -- sec min hr day mon yr ; 10.6.2.2292 
\ *G The ANS Forth word to return time and date.
  lseconds @
  secs/min /umod
  mins/hr /umod
  hrs/day /umod				\ -- sec min hr day#
  u>jday >date				\ -- sec min hr day mon yr
;

: dow		\ -- dow ; 0=Sunday
\ *G Return the day-of-week value last read from the RTC.
\ ** Sunday returns 0, Monday returns 1 and so on.
  lseconds @ 0 secs/day um/mod nip	\ -- uday
  u>jday weekday
;

: settime&date	\ sec min hr day mon yr --
\ *G Use the given values for time and date, and set LSECONDS
\ ** and any underlying RTC hardware.
  date> j>uday secs/day *		\ -- sec min hr lsecs
  swap secs/hr * +
  swap secs/min * +
  +
  lseconds !
;
  

\ ************************
\ *S Time and date display
\ ************************
\ *P These functions rely on the ANS Forth word TIME&DATE
\ ** ( -- s m h dd mm yy ) and the non-standard DOW
\ ** ( -- dow, 0=Sun) to get the day of the week.

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

: .dow		\ --
\ *G Display day of week.
  dow 3 * days$ 1+ + 3 type
;

create months	\ -- addr
\ *G String containing 3 character text for the months.
  z", JanFebMarAprMayJunJulAugSepOctNovDec"

: .2r		\ n --
\ *G Display n as a two digit number with leading zeros.
  s>d <# # # #> type
;

: .4r		\ n --
\ *G Display n as a four digit number with leading zeros.
  s>d <# # # # # #> type
;

: .Time/Date	\ --
\ *G Display the system time
\ ** The format is:
\ *C   hh:mm:ss dd Mmm yyyy
  time&date				\ -- s m h dd mm yy
  dup #100 <				\ correct two digit year clocks
  if  #2000 +  endif
  rot .2r space				\ day - generate date
  swap 1- 3 * months + 3 type space	\ month
  .4r space				\ year
  .2r ." :"			 	\ hours - generate time 
  .2r ." :"				\ minutes
  .2r					\ seconds
;

: .zone		\ zone --
  if  3 spaces  else  ." GMT"  endif
;

: .AnsiDate	\ zone --
\ *G Display the day of week, time and date. If zone is 0
\ ** GMT is displayed. The format is Internet time:
\ *C   "dow, hh:mm:ss dd Mmm yyyy [GMT] "
  .dow ." , " .time/date space .zone space
;


\ ======
\ *> ###
\ ======

