\ ANS memory management extensions for 32 bit targets

\ 20000321 SFP003 Fixed more signed/unsigned issues. These are all benign
\                 in this code, as the maximum allocation is 0x00FFFFFF,
\		  but changing it makes the code better to read and maintain.
\ 19990315 SJC001 Fixed sign bug in size. Must treat the requested size as
\                 unsigned in ALLOCATE. 
\ 19980803 MSD001 Correct the returned address when RESIZE fails.  The ANS spec
\          MSD001 says that in this situation, RESIZE should return a
\          MSD001 (non-zero) return code and the *original* block pointer
\          MSD001 address.  This code used to return the block pointer of the
\          MSD001 failed ALLOCATE when the block was being increased in size.
\          MSD001 Since ALLOCATE returns an undefined value, RESIZE *cannot*
\          MSD001 use it as the block size to return when trying to increase
\          MSD001 the block size and ALLOCATE fails.

\ we use 2 equates during compilation
\ to allocate a contiguous block of RAM for the heap
\ STARTOFHEAP is funnily enough the start of the heap
\ SIZEOFHEAP is the max size of RAM available for the heap
\
\ the heap is controlled using a single cell (32 bits) per block
\ this cell is used as two parts:
\
\ bits 31-24 = flag
\ bits 23-0 = #bytes
\
\ the flag indicates the status of the block:
\
\       EE = End
\       FF = Free
\       AA = Allocated
\       others may be added later for type management

\ the heap MUST be initialised before use by calling init-heap

\ status returned is 0 for OK <> 0 is fail code


only forth definitions
decimal

[defined] environment [if]
  also environment definitions
  TRUE constant MEMORY-ALLOC
  previous definitions
[then]

\ ------------------------------------------------------------------------------

$00FFFFFF       equ MASKSIZE
$FF000000       equ MASKTYPE

$EE000000       equ LASTBLOCK                   \ marker for end of heap
$FF000000       equ FREEBLOCK                   \ marker for free block on heap
$AA000000       equ ALLOCBLOCK                  \ marker for allocated block

\ The following constant is the minimum *total* size occupied by a block. This
\ size includes overheads generated by alignment retrictions as well as those
\ incurred by the fact that the size and status of a block have to be stored
\ somewhere.
1 cells equ /HEAPINFO				\ size of heap information
1 aligned /HEAPINFO + equ MINIMUMHEAPENTRYSIZE	\ minimum total block size

\ ------------------------------------------------------------------------------

sizeofheap buffer: STARTOFHEAP


\ ------------------------------------------------------------------------------
\ INIT-HEAP
\ The heap is initialised by creating 2 blocks.
\ Block1 starts at the beginning and is marked as a free block
\ Block2 Is a null marker at the end of heap space
\ ------------------------------------------------------------------------------

: init-heap	\ --
  startofheap sizeofheap erase                  \ erase heap memory

  sizeofheap /HEAPINFO 2* -			\ calculate free block size
  MASKSIZE and  FREEBLOCK or			\ mask in FREEBLOCK indicator
  startofheap !                                 \ write header

  LASTBLOCK					\ end of heap marker
  startofheap sizeofheap + /HEAPINFO - !	\ to last block in heap
;

\ ------------------------------------------------------------------------------
\ GARBAGE-COLLECTION
\ ------------------------------------------------------------------------------

-short-branches
: compact	{ | *curr *next -- }

  startofheap -> *curr                          \ ptr to current block

  begin

    \ SKIP PAST ANY ALLOCd BLOCKS
    begin
      *curr @ MASKTYPE and ALLOCBLOCK =
     while
      *curr @ MASKSIZE and /HEAPINFO +		\ block size
      *curr + -> *curr				\ move *curr to next block
    repeat

    *curr @ MASKTYPE and FREEBLOCK =            \ are we on a free block?
   while
    \ SCAN for a new non-free block
    *curr -> *next
    begin
      *next @ MASKTYPE and FREEBLOCK =		\ currblock FREE?
     while
      *next @ MASKSIZE and /HEAPINFO +		\ yep, move on to next
      *next + -> *next
    repeat

    \ *curr = current free block
    \ *next = ptr the the next non free block
    *next *curr - /HEAPINFO -			\ calc size of new free block
    MASKSIZE and FREEBLOCK or
    *curr !					\ and store it
    
    *curr @ MASKSIZE and /HEAPINFO +
    *curr + -> *curr       			\ here we go again
  repeat
;
+short-branches

\ ------------------------------------------------------------------------------
\ ALLOCATE
\ Attempt to allocate some memory from heap
\ Walk heap looking for a single block big enough to cope
\ If block is more than 2 CELLS bigger than required split it
\ Allocate 1st part (or all) of free block
\ ------------------------------------------------------------------------------

-short-branches
: allocate	{ #bytes | *curr cbt cbs *blk -- addr status }

  startofheap -> *curr
  #bytes aligned -> #bytes

\ STEP#1: Find a big enough free block and set *BLK (-1 means failed)

  -1 -> *blk                                    \ initial ptr/rc
  begin
    *curr @ MASKSIZE AND -> cbs                 \ curr block size
    *curr @ MASKTYPE AND -> cbt                 \ curr block type

\    cbs #bytes >=                                                              \ sjc001
     #bytes cbs u<                                                              \ sjc001
     cbt FREEBLOCK =
    and                                         \ is block free and big enough?
    if *curr -> *blk then                       \ yep! Set Pointer

    *curr cbs CELL+ + -> *curr                  \ point curr at next block
  cbt LASTBLOCK =                               \ continue until LASTBLOCK
  *blk -1 <> or                                 \ .. or a valid block is found
  until

\ Here we have *BLK pointing to a block (or error code -1) and CBS as it's size

  *blk -1 <>
  if
    cbs #bytes 2 CELLS + u>                     \ split block?
    if
       \ Size of new free block (top section of *blk) is current freesize -CELL
       \ since I now need a new control CELL

       cbs #bytes CELL+ -                       \ calc new free block size
       MASKSIZE and FREEBLOCK or                \ mask in FREEBLOCK indicator
       *blk #bytes + /HEAPINFO +		\ addr of second block control
       !                                        \ flag it!

       #bytes                                   \ new size for current block
       *blk !                                   \ and store it

    then
    *blk @ MASKSIZE and ALLOCBLOCK or *blk !    \ store ALLOC type in block

    *blk /HEAPINFO +				\ return data addr
    0                                           \ and 0 for GOOD status

  else
    0 -1                                        \ failure return codes
  then
;
+short-branches


\ ------------------------------------------------------------------------------
\ FREE
\ Attempt to free a heap block
\ ------------------------------------------------------------------------------

: free		\ *ptr -- status
  /HEAPINFO - dup @ MASKTYPE AND ALLOCBLOCK =	\ if block was alloc'd
  if
    dup @  MASKSIZE and  FREEBLOCK or  swap !	\ reset type to FREE
    0                                           \ GOOD rc
    compact
  else
    drop  -1					\ failed rc!
  then
;

\ ------------------------------------------------------------------------------
\
\ RESIZE
\
\ Try and re-size an allocated block to a new size.  (In the following
\ comments, "new size" refers to the new size requested by the user *after*
\ alignment requirements have been taken into consideration.)
\
\ There are three different cases that we have to consider:
\
\       The new size is less than or equal to the old size minus the minimum
\       size taken by a heap entry.  In this case, we shrink the block size
\       making the free space into a free block.  We return the *existing*
\       block pointer because the user's data hasn't been moved.
\
\       The new size is *greater* than the old block size.  Since the user
\       has requested more space than is available in the current block, we
\       have to allocate a *new* block, copy the user's data into it and
\       then free the *old* block.  We return the pointer to the
\       newly-allocated block, of course.
\
\       If neither of the above cases apply, then we just return the
\       existing pointer because there is not enough un-used space in the
\       block to create a new (free) heap block.
\
\ ------------------------------------------------------------------------------

-short-branches
: resize	{ *ptr newsize | currsize -- *newptr ior }

  newsize aligned -> newsize                    \ align new size

  *ptr cell- @ MASKSIZE and -> currsize         \ get current block size

  newsize  currsize MINIMUMHEAPENTRYSIZE - u> 0=  \ Will we have a block's worth
  if                                            \ of free space after resizing?
    newsize MASKSIZE and                        \ mask newsize
    ALLOCBLOCK or *ptr cell- !                  \ and store as size of curblk
    currsize newsize - cell -                   \ calc new free size for 2nd
    MASKSIZE and FREEBLOCK or                   \ mask it up
    *ptr newsize +                              \ ptr to new block
    !                                           \ store it
    *ptr 0                                      \ rc's
  else
    newsize currsize u>                         \ Must we allocate a bigger
    if                                          \ block?
      newsize allocate 0=                       \ get a new block
      if
        dup >r                                  \ preserve new ptr
        *ptr swap currsize move                 \ move existing data to new blk
        *ptr free drop                          \ free old block now
        r> 0                                    \ flag as ok and cool!
      else
        drop *ptr -1                            \ failed to alloc new big block \ MSD001
      then
    else                                        \ No space to generate a new
      *ptr 0                                    \ free block - use current block.
    then
  then
;
+short-branches


\ ------------------------------------------------------------------------------
\ SIZE
\ Return the size of an allocated block or -1 if there's an error.
\ ------------------------------------------------------------------------------

: size		\ *ptr -- currsize | -1
  /HEAPINFO - @ dup                           \ control control
  MASKTYPE and  ALLOCBLOCK =
  if
    MASKSIZE and
  else
    drop -1
  then
;


\ ------------------------------------------------------------------------------
\ .HEAP
\ Walk the heap displaying block information
\ ------------------------------------------------------------------------------

: .heap#	( x -- ; Print 32bit HEX number )
  base @ >r
  hex
  s>d <# # # # # # # # # #> type [char] h emit
  r> base !
;

-short-branches
: .heap		{ | *curr freecnt freesz usedcnt usedsz -- }

  cr
  ." Heap is " sizeofheap . ."  bytes long, starting at address " startofheap .heap#

  cr
  startofheap -> *curr
  0 -> freecnt
  0 -> freesz
  0 -> usedcnt
  0 -> usedsz

  cr ." BlockAddr   BlockType   BlockSize "

  begin
    *curr @ MASKSIZE and                        \ get block size and type
    *curr @ MASKTYPE and
    dup LASTBLOCK <>                            \ if not at end
    if
      dup FREEBLOCK =
      if
        over freesz + -> freesz
        freecnt 1+ -> freecnt
        cr *curr CELL+ .heap# ."       FREE     " over .heap#
      then
      dup ALLOCBLOCK =
      if
        over usedsz + -> usedsz
        usedcnt 1+ -> usedcnt
        cr *curr CELL+ .heap# ."       USED     " over .heap#
      then
    then
    swap *curr + cell+ -> *curr                 \ point *curr to next block
  LASTBLOCK =                                   \ finished when hit LASTBLOCK
  until

  cr
  cr freesz . ."  bytes free in " freecnt . ."  blocks"
  cr usedsz . ."  bytes used in " usedcnt . ."  blocks"
;
+short-branches


\ ------------------------------------------------------------------------------
\ HEAPOK?
\ Walk the heap and return TRUE if the heap is "well"
\ ------------------------------------------------------------------------------

: heapok?	( -- t/f )

  startofheap

  begin

    dup startofheap sizeofheap + cell- u>
    if
      cr ." Past end of heap at " .heap#
      false exit
    then

    dup @ LASTBLOCK =
    if
      startofheap sizeofheap + cell- =  \ Last block must be at *end* of heap.
      dup not
      if
        cr ." Misplaced end-of-heap marker"
      endif
      exit
    then

    dup @ MASKTYPE and
    dup FREEBLOCK =
    swap ALLOCBLOCK =
    or not
    if
      cr ." Bad heap marker at " .heap#
      drop false exit
    then

    dup @ MASKSIZE and + cell+          \ Point to next block.
  again
;

\ ------------------------------------------------------------------------------



