\ ANS memory management extensions for 32 bit targets
\ Use HEAP16.FTH for the documentation.

((
Copyright (c) 2000-2003
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeltd.demon.co.uk
     tech-support@mpeltd.demon.co.uk
web: www.mpeltd.demon.co.uk

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====
De-local and Forthify.

Change History
==============
20030331 SFP006 Added semaphore locking option.
20030315 SFP005 Updated ALLOCATE and took advantage of signed
		size arithmetic again for better code on some CPUs.
20030312 SFP004 Revamped .HEAP to ease leak checking.
20000321 SFP003 Fixed more signed/unsigned issues. These are all benign
                in this code, as the maximum allocation is 0x00FFFFFF,
                but changing it makes the code better to read and maintain.
19990315 SJC001 Fixed sign bug in size. Must treat the requested size as
                unsigned in ALLOCATE. 
19980803 MSD001 Correct the returned address when RESIZE fails.  The ANS spec
         MSD001 says that in this situation, RESIZE should return a
         MSD001 (non-zero) return code and the *original* block pointer
         MSD001 address.  This code used to return the block pointer of the
         MSD001 failed ALLOCATE when the block was being increased in size.
         MSD001 Since ALLOCATE returns an undefined value, RESIZE *cannot*
         MSD001 use it as the block size to return when trying to increase
         MSD001 the block size and ALLOCATE fails.

we use 2 equates during compilation
to allocate a contiguous block of RAM for the heap
 STARTOFHEAP is the start address of the heap
 SIZEOFHEAP is the max size of RAM available for the heap

the heap is controlled using a single cell (32 bits) per block
this cell is used as two parts:

 bits 31-24 = flag
 bits 23-0 = #bytes

the flag indicates the status of the block:
      EE = End
      FF = Free
      AA = Allocated
      others may be added later for type management

the heap MUST be initialised before use by calling init-heap

status returned is 0 for OK, <> 0 is a fail code
))

only forth definitions
decimal

[defined] environment [if]
  also environment definitions
  TRUE constant MEMORY-ALLOC
  previous definitions
[then]

\ ------------------------------------------------------------------------------


\ If you use a pre-emptive scheduler or need to use the heap
\ routines inside interrupt routines, you must define suitable
\ heap lock and unlock routines and set the equate LOCKHEAP? to
\ non-zero.
\ *D LockHeap=0 no heap locking
\ *D LockHeap=1 heap locking by turning off interrupts
\ *D LockHeap=2 heap locking by semaphore.

0 equ LockHeap?	\ -- n

LockHeap? 0 = [if]	\ no heap locking
compiler
: InitHeapLock	;	\ --
: LockHeap	;	\ -- 
: UnlockHeap	;	\ --
target
[then]

LockHeap? 1 = [if]	\ interrupt locking
compiler
: InitHeapLock	;	\ --
: LockHeap	[i  ;	\ -- ; R: -- x
: UnlockHeap	i]  ;	\ -- ; R: x --
target
[then]

LockHeap? 2 = [if]	\ semaphore locking
Semaphore Heap-Access
compiler
: InitHeapLock	0 1 Heap-Access 2!  ;	\ --
: LockHeap	Heap-Access request  ;	\ --
: UnlockHeap	Heap-Access signal  ;	\ --
target
[then]


$00FFFFFF equ MASKSIZE
$FF000000 equ MASKTYPE

$EE000000 equ LASTBLOCK			\ marker for end of heap
$FF000000 equ FREEBLOCK                 \ marker for free block on heap
$AA000000 equ ALLOCBLOCK                \ marker for allocated block

\ The following constant is the minimum *total* size occupied by a block. This
\ size includes overheads generated by alignment restrictions as well as those
\ incurred by the fact that the size and status of a block have to be stored
\ somewhere.
1 cells equ /HEAPINFO				\ size of heap information
1 aligned /HEAPINFO + equ MINIMUMHEAPENTRYSIZE	\ minimum total block size

\ ------------------------------------------------------------------------------

sizeofheap buffer: STARTOFHEAP


\ ------------------------------------------------------------------------------
\ INIT-HEAP
\ The heap is initialised by creating 2 blocks.
\ Block1 starts at the beginning and is marked as a free block
\ Block2 Is a null marker at the end of heap space
\ ------------------------------------------------------------------------------

: init-heap	\ --
  InitHeapLock

  startofheap sizeofheap erase          \ erase heap memory

  sizeofheap /HEAPINFO 2* -		\ calculate free block size
  MASKSIZE and  FREEBLOCK or		\ mask in FREEBLOCK indicator
  startofheap !                         \ write header

  LASTBLOCK				\ end of heap marker
  startofheap sizeofheap + /HEAPINFO - ! \ to last block in heap
;

\ ------------------------------------------------------------------------------
\ GARBAGE-COLLECTION
\ ------------------------------------------------------------------------------

-short-branches

: compact	\ --
  startofheap
  begin
    \ skip past any allocated blocks
    begin				\ -- *curr
      dup @ MASKTYPE and ALLOCBLOCK =
     while
      dup @ MASKSIZE and + /HEAPINFO +	\ step to next block
    repeat
    dup @ MASKTYPE and FREEBLOCK =	\ are we on a free block?
   while	
    \ scan for a new non-free block
    dup
    begin				\ -- *curr *next
      dup @ MASKSIZE and + /HEAPINFO +	\ move on to next block
      dup @ MASKTYPE and FREEBLOCK <>	\ until not FREE
    until				\ -- *curr *next'
    \ *curr = current free block
    \ *next = next non free block
    tuck				\ preserve *next for next iteration
    over - /HEAPINFO -			\ -- *next *curr sizefree
    MASKSIZE and FREEBLOCK or		\ mark as free block
    swap !
  repeat
  drop
;

+short-branches


\ ------------------------------------------------------------------------------
\ ALLOCATE
\ Attempt to allocate some memory from heap
\ Walk heap looking for a single block big enough to cope
\ If block is more than 2 CELLS bigger than required split it
\ Allocate 1st part (or all) of free block
\ ------------------------------------------------------------------------------

-short-branches

: SplitBlk	\ size *blk -- ; split block, mark top as free
  over 2 cells +			\ split above this size
  over @ MASKSIZE and			\ current block size
  u< if					\ split?
     \ Size of new free block (top section of *blk) is current
     \ freesize -CELL since we now need a new control CELL
     2dup @ MASKSIZE and swap cell+ -	\ new free block size
     FREEBLOCK or			\ -- size *blk sizef ; mask in FREEBLOCK indicator
     2 pick 2 pick + /HEAPINFO + !	\ -- size *blk ; set second block control info
     swap ALLOCBLOCK or  swap !		\ set new control info
  else
     nip				\ discard size
     dup @ MASKSIZE and			\ use existing size
     ALLOCBLOCK or  swap !		\ mark allocated
  then
;

: allocate	\ #bytes -- addr status
  aligned dup MASKSIZE u>		\ check #bytes, we can use
  if  -1  exit  endif			\ signed arithmetic on #bytes
  \ -- #bytes
\ Find a big enough free block, fail at LASTBLOCK
  LockHeap				\ SFPxxx
  startofheap				\ -- #bytes start
  begin
    dup @ MASKTYPE and LASTBLOCK <>	\ continue while not LASTBLOCK
   while				\ -- #bytes *blk
    dup @ MASKTYPE and FREEBLOCK = if	\ if block is free
      2dup @ MASKSIZE and <= if		\ and big enough
        tuck SplitBlk			\ split the block to alloc+free
	/HEAPINFO +  0 			\ return pointer and 0
	UnlockHeap  exit		\ SFPxxx
      then
    then
    dup @ MASKSIZE and + cell+		\ step to next block
  repeat
  2drop  0 -1
  UnlockHeap				\ SFPxxx
;
+short-branches


\ ------------------------------------------------------------------------------
\ FREE
\ Attempt to free a heap block
\ ------------------------------------------------------------------------------

: free		\ *ptr -- status
  LockHeap				\ SFPxxx
  /HEAPINFO -				\ step back to control block
  dup @ MASKTYPE and ALLOCBLOCK = if	\ if block was allocated
    dup @
    MASKSIZE and  FREEBLOCK or		\ reset type to FREE
    swap !
    0					\ GOOD
    compact
  else
    drop  -1				\ failed
  then
  UnlockHeap				\ SFPxxx
;

\ ------------------------------------------------------------------------------
\ 
\ RESIZE
\ 
\ Try and re-size an allocated block to a new size.  (In the following
\ comments, "new size" refers to the new size requested by the user *after*
\ alignment requirements have been taken into consideration.)
\ 
\ There are three different cases that we have to consider:
\ 
\       The new size is less than or equal to the old size minus the minimum
\       size taken by a heap entry.  In this case, we shrink the block size
\       making the free space into a free block.  We return the *existing*
\       block pointer because the user's data hasn't been moved.
\ 
\       The new size is *greater* than the old block size.  Since the user
\       has requested more space than is available in the current block, we
\       have to allocate a *new* block, copy the user's data into it and
\       then free the *old* block.  We return the pointer to the
\       newly-allocated block, of course.
\
\       If neither of the above cases apply, then we just return the
\       existing pointer because there is not enough un-used space in the
\       block to create a new (free) heap block.
\
\ ------------------------------------------------------------------------------

-short-branches
\ works, ARM=396 bytes
: resize	{ *ptr newsize | currsize -- *newptr ior }
  newsize aligned -> newsize            \ align new size
  *ptr cell- @ MASKSIZE and -> currsize \ get current block size
  newsize  currsize MINIMUMHEAPENTRYSIZE - u> 0=  \ Will we have a block's worth
  if                                    \ of free space after resizing?
    LockHeap				\ SFP006
    newsize MASKSIZE and                \ mask newsize
    ALLOCBLOCK or *ptr cell- !          \ and store as size of curblk
    currsize newsize - cell -           \ calc new free size for 2nd
    MASKSIZE and FREEBLOCK or           \ mask it up
    *ptr newsize +                      \ ptr to new block
    !                                   \ store it
    *ptr 0                              \ rc's
    UnlockHeap  exit			\ SFP006
  endif
  newsize currsize u> if		\ Must we allocate bigger block
    newsize allocate 0= if              \ get a new block
      dup >r                            \ preserve new ptr
      *ptr swap currsize move           \ move existing data to new blk
      *ptr free drop                    \ free old block now
      r> 0                              \ flag as ok
    else
      drop *ptr -1                      \ failed to alloc new big block \ MSD001
    then
  else                                  \ No space to generate a new
    *ptr 0                              \ free block - use current block.
  then
;
(( works: original
: resize	{ *ptr newsize | currsize -- *newptr ior }

  LockHeap				\ SFPxxx

  newsize aligned -> newsize            \ align new size

  *ptr cell- @ MASKSIZE and -> currsize \ get current block size

  newsize  currsize MINIMUMHEAPENTRYSIZE - u> 0=  \ Will we have a block's worth
  if                                    \ of free space after resizing?
    newsize MASKSIZE and                \ mask newsize
    ALLOCBLOCK or *ptr cell- !          \ and store as size of curblk
    currsize newsize - cell -           \ calc new free size for 2nd
    MASKSIZE and FREEBLOCK or           \ mask it up
    *ptr newsize +                      \ ptr to new block
    !                                   \ store it
    *ptr 0                              \ rc's
  else
    newsize currsize u>                 \ Must we allocate a bigger
    if                                  \ block?
      newsize allocate 0=               \ get a new block
      if
        dup >r                          \ preserve new ptr
        *ptr swap currsize move         \ move existing data to new blk
        *ptr free drop                  \ free old block now
        r> 0                            \ flag as ok
      else
        drop *ptr -1                    \ failed to alloc new big block \ MSD001
      then
    else                                \ No space to generate a new
      *ptr 0                            \ free block - use current block.
    then
  then
  UnlockHeap				\ SFPxxx
;
))
(( In progress
: resize	\ *ptr newsize -- *newptr ior
  aligned dup MASKSIZE u> if		\ align and check size
    drop -1  exit
  endif
  LockHeap				\ SFPxxx
  swap /HEAPINFO -			\ -- newsize *blk
  dup @ MASKSIZE and >r			\ get current size
  over r@ u< if				\ new < current?
    tuck SplitBlk
    /HEAPINFO + 0
    r> drop  unlockHeap  exit		\ SFPxxx
  then
  over r@ = if				\ new = current?
    nip  /HEAPINFO + 0
    r> drop  UnlockHeap  exit		\ SFPxxx
  then					\ -- newsize *blk
  \ larger
  swap allocate if			\ -- *blk addr
    drop
    /HEAPINFO +  -1			\ failed
    r> drop  UnlockHeap  exit		\ SFPxxx
  then
  swap /HEAPINFO + 2dup			\ -- addr *data addr *data
  swap r> move				\ -- addr *data ; copy data
  free drop				\ free old block
  0
  UnlockHeap
;
))
+short-branches


\ ------------------------------------------------------------------------------
\ SIZE
\ Return the size of an allocated block or -1 if there's an error.
\ ------------------------------------------------------------------------------

: size		\ *ptr -- currsize|-1
\  LockHeap				\ SFPxxx
  /HEAPINFO -				\ control block
  @ dup  MASKTYPE and  ALLOCBLOCK = if	\ allocated?
    MASKSIZE and			\ return size
  else
    drop -1
  then
\  UnlockHeap				\ SFPxxx
;


heap-diags? [if]

\ ------------------------------------------------------------------------------
\ .HEAP
\ Walk the heap displaying block information
\ ------------------------------------------------------------------------------

: .heap#	( x -- ; Print 32bit HEX number )
  base @ >r
  hex
  s>d <# # # # # # # # # #> type [char] h emit
  r> base !
;

-short-branches
: .heap		{ | *curr freecnt freesz usedcnt usedsz -- ; SFP004 }
  cr
  ." Heap size is " sizeofheap . ." bytes at " startofheap .heap#
  cr
  startofheap -> *curr
  0 -> freecnt
  0 -> freesz
  0 -> usedcnt
  0 -> usedsz

  cr ." BlockAddr   BlockType   BlockSize "
  begin
    *curr @ MASKSIZE and                \ get block size and type
    *curr @ MASKTYPE and		\ -- size type
    case  dup
      LASTBLOCK of
        over CELL + usedsz + -> usedsz
        cr  *curr CELL+ .heap#  ."       LAST     " over .heap#
      endof
      FREEBLOCK of
        over CELL + freesz + -> freesz
        freecnt 1+ -> freecnt
        cr  *curr CELL+ .heap#  ."       FREE     " over .heap#
      endof
      ALLOCBLOCK of
        over CELL + usedsz + -> usedsz
        usedcnt 1+ -> usedcnt
        cr  *curr CELL+ .heap#  ."       USED     " over .heap#
      endof
        cr  *curr CELL+ .heap#  ."       BAD!     " over .heap#
    endcase
    swap *curr + cell+ -> *curr         \ point *curr to next block
    LASTBLOCK =				\ finished when hit LASTBLOCK
  until

  cr
  cr freesz . ." bytes free in " freecnt . ." blocks"
  cr usedsz . ." bytes used in " usedcnt . ." blocks"
  cr freesz usedsz + . ." bytes total"
  cr sizeofheap . ." is SIZEOFHEAP"
;
+short-branches


\ ------------------------------------------------------------------------------
\ HEAPOK?
\ Walk the heap and return TRUE if the heap is "well"
\ ------------------------------------------------------------------------------

: heapok?	\ -- t/f
  startofheap
  begin
    dup startofheap sizeofheap + cell- u> if
      cr ." Past end of heap at " .heap#
      false exit
    then
    dup @ LASTBLOCK = if
      startofheap sizeofheap + cell- =  \ Last block must be at *end* of heap.
      dup not if
        cr ." Misplaced end-of-heap marker"
      endif
      exit
    then
    dup @ MASKTYPE and
    dup FREEBLOCK =
    swap ALLOCBLOCK =
    or not if
      cr ." Bad heap marker at " .heap#
      drop false exit
    then
    dup @ MASKSIZE and + cell+          \ Point to next block.
  again
;

\ ------------------------------------------------------------------------------

[then]


